#include "stdafx.h"

#include "../system.h"
#include "../Core.h"
#include "MIPS.h"
#include "MIPSInt.h"
#include "MIPSTables.h"
#include "MIPSVFPUUtils.h"

#include "../PSP/PSPFirmwareHLE.h"

#include <limits>

#define R(i)   (currentMIPS->r[i])
#define RF(i)  (*(float*)(&(currentMIPS->r[i])))
#define V(i)   (currentMIPS->v[i])
#define VI(i)   (*(u32*)(&(currentMIPS->v[i])))
#define FI(i)  (*(u32*)(&(currentMIPS->f[i])))
#define FsI(i) (*(s32*)(&(currentMIPS->f[i])))
#define PC     (currentMIPS->pc)

#define _RS   ((op>>21) & 0x1F)
#define _RT   ((op>>16) & 0x1F)
#define _RD   ((op>>11) & 0x1F)
#define _FS   ((op>>11) & 0x1F)
#define _FT   ((op>>16) & 0x1F)
#define _FD   ((op>>6 ) & 0x1F)
#define _POS  ((op>>6 ) & 0x1F)
#define _SIZE ((op>>11 ) & 0x1F)

#define HI currentMIPS->hi
#define LO currentMIPS->lo

static u32 nextPC = 0;
static bool inDelaySlot = 0;




void ApplyPrefixST(float *v, u32 data, DataSize size)
{
	int n = GetNumElements(size);
	float origV[4];
	static const float constantArray[8] = {0.f,1.f,2.f,0.5f,3.f,1.f/3.f,0.25f,1.f/6.f};

	for (int i=0; i<n; i++)
	{
		origV[i] = v[i];
	}

	for (int i=0; i<n; i++)
	{
		int regnum = (data>>(i*2)) & 3;
		int abs    = (data>>(8+i)) & 1;
		int negate = (data>>(16+i)) & 1;
		int constants = (data>>(12+i)) & 1;

		if (!constants)
		{
			v[i] = origV[regnum];
			if (abs)
				v[i] = fabs(v[i]);
		}
		else
		{
			v[i] = constantArray[regnum + (abs<<2)];
		}

		if (negate)
			v[i] = -v[i];
	}
}

void ApplySwizzleS(float *v, DataSize size)
{
	ApplyPrefixST(v, currentMIPS->vfpuCtrl[VFPU_CTRL_SPREFIX], size);
}

void ApplySwizzleT(float *v, DataSize size)
{
	ApplyPrefixST(v, currentMIPS->vfpuCtrl[VFPU_CTRL_TPREFIX], size);
}


void ApplyPrefixD(float *v, DataSize size)
{
	int n = GetNumElements(size);
	bool writeMask[4];
	u32 data = currentMIPS->vfpuCtrl[VFPU_CTRL_DPREFIX];
	for (int i=0; i<n; i++)
	{
		int sat = (data>>i*2)&3;
		int mask = (data>>(8+i))&1;
		writeMask[i] = mask ? true : false;
		if (sat == 1)
		{
			if (v[i] > 1.0f) v[i]=1.0f;
			if (v[i] < 0.0f) v[i]=0.0f;
		}
		else if (sat == 3)
		{
			if (v[i] > 1.0f)  v[i]=1.0f;
			if (v[i] < -1.0f) v[i]=-1.0f;
		}
	}
	//ApplySwizzle(v, currentMIPS->vfpuCtrl[1], GetNumElements(size));
}

void EatPrefixes()
{
	bool noWriteMask[4] = {false,false,false,false};
	currentMIPS->vfpuCtrl[VFPU_CTRL_SPREFIX]=0xe4; //passthru
	currentMIPS->vfpuCtrl[VFPU_CTRL_TPREFIX]=0xe4; //passthru
	currentMIPS->vfpuCtrl[VFPU_CTRL_DPREFIX]=0;
	SetWriteMask(noWriteMask);
}




namespace MIPSInt
{
#define S_not(a,b,c) (a<<2)|(b)|(c<<5)
#define SgetA(v) (((v)>>2)&0x7)
#define SgetB(v) ((v)&3)
#define SgetC(v) (((v)>>5)&0x3)
#define VS(m,row,col) V(m*4+(row)+(col)*32)

	void Int_VPFX(u32 op)
	{
		int data = op & 0xFFFFF;
		int regnum = (op>>24) & 3;
		currentMIPS->vfpuCtrl[VFPU_CTRL_SPREFIX+regnum] = data;
		PC+=4;
	}

	void Int_SVQ(u32 op)
	{
		int imm = (signed short)(op&0xFFFC);
		int rs = _RS;
		int vt = (((op>>16)&0x1f))|((op&1)<<5);

		u32 addr = R(rs) + imm;

		switch (op>>26)
		{
		case 53: //lvl.q/lvr.q
			if ((op&2) == 0)
				break;
		case 54: //lv.q
			WriteVector((const float*)GetMemPointer(addr), V_Quad, vt);
			break;
		case 61://temporary solution
			if ((op&2) == 0)
				break;
		case 62: //sv.q
			ReadVector((float*)GetMemPointer(addr), V_Quad, vt);
			break;
		default:
			_dbg_assert_msg_(CPU,0,"Trying to interpret VQ instruction that can't be interpreted");
			break;
		}
		PC += 4;
	}

	

	void Int_VMatrixInit(u32 op)
	{
		static const float idt[16] = 
		{
			1,0,0,0,
			0,1,0,0,
			0,0,1,0,
			0,0,0,1
		};
		static const float zero[16] = 
		{
			0,0,0,0,
			0,0,0,0,
			0,0,0,0,
			0,0,0,0
		};
		static const float one[16] = 
		{
			1,1,1,1,
			1,1,1,1,
			1,1,1,1,
			1,1,1,1
		};
		int vd = _VD;
		DataSize sz = GetMtxSize(op);
		const float *m;

		switch ((op>>16) & 0xF)
		{
		case 3: m=idt; break; //identity
		case 6: m=zero; break;
		case 7: m=one; break;
		default:
			_dbg_assert_msg_(CPU,0,"Trying to interpret instruction that can't be interpreted");
			break;
		}

		WriteMatrix(m, sz, 4, vd);
		PC += 4;
		EatPrefixes();
	}

	void Int_VVectorInit(u32 op)
	{
		int vd = _VD;
		DataSize sz = GetVecSize(op);
		int n = GetNumElements(sz);
		const float ones[4]={1,1,1,1};
		const float zeros[4]={0,0,0,0};
		const float *v;
		switch ((op>>16) & 0xF)
		{
		case 6: v=zeros; break;
		case 7: v=ones; break;
		default:
			_dbg_assert_msg_(CPU,0,"Trying to interpret instruction that can't be interpreted");
			break;
		}
		float o[4];
		for (int i=0; i<n; i++)
			o[i] = v[i];
		ApplyPrefixD(o,sz);
		WriteVector(o,sz,vd);

		EatPrefixes();
		PC += 4;
	}

	void Int_Viim(u32 op)
	{
		int vt = _VT;
		int imm = op&0xFFFF;
		//V(vt) = (float)imm;
		int type = (op >> 23) & 7;
		if (type == 6)
			V(vt) = (float)imm;
		else if (type == 7)
			V(vt) = Float16ToFloat32((u16)imm);
		else
			_dbg_assert_msg_(CPU,0,"Trying to interpret instruction that can't be interpreted");
		
		PC+=4;
		EatPrefixes();
	}



	void Int_Vidt(u32 op)
	{
		int vd = _VD;
		DataSize sz = GetVecSize(op);
		float f[4];
		switch (sz)
		{
		case V_Pair:
			f[0] = (vd&1)==0 ? 1.0f : 0.0f;
			f[1] = (vd&1)==1 ? 1.0f : 0.0f;
			break;
		case V_Quad:
			f[0] = (vd&3)==0 ? 1.0f : 0.0f;
			f[1] = (vd&3)==1 ? 1.0f : 0.0f;
			f[2] = (vd&3)==2 ? 1.0f : 0.0f;
			f[3] = (vd&3)==3 ? 1.0f : 0.0f;
			break;
		default:
			_dbg_assert_msg_(CPU,0,"Trying to interpret instruction that can't be interpreted");
			break;
		}
		WriteVector(f, sz, vd);
		PC+=4;
		EatPrefixes();
	}

	void Int_Vmmul(u32 op)
	{
		float s[16];
		float t[16];
		float d[16];

		int vd = _VD;
		int vs = _VS;
		int vt = _VT;
		DataSize sz = GetMtxSize(op);
		int n = GetMatrixSide(sz);

		ReadMatrix(s,sz,4,vs);
		ReadMatrix(t,sz,4,vt);

		for (int a = 0; a<n; a++)
		{
			for (int b = 0; b<n; b++)
			{
				float sum=0;
				for (int c=0; c<n; c++)
				{
					sum += s[b*4+c] * t[a*4+c];
				}
				d[a*4+b]=sum;
			}
		}

		WriteMatrix(d,sz,4,vd);
		PC+=4;
		EatPrefixes();
	}

	void Int_Vmmov(u32 op)
	{
		float s[16];
		int vd = _VD;
		int vs = _VS;
		DataSize sz = GetMtxSize(op);
		ReadMatrix(s,sz,4,vs);
		WriteMatrix(s,sz,4,vd);
		PC+=4;
		EatPrefixes();
	}

	void Int_Vflush(u32 op)
	{
		LOG(HLE,"vflush");
		PC+=4;
	}

	void Int_VV2Op(u32 op)
	{
		float s[4], d[4];
		int vd = _VD;
		int vs = _VS;
		DataSize sz = GetVecSize(op);
		ReadVector(s,sz,vs);
		ApplySwizzleS(s,sz);
		for (int i=0; i<GetNumElements(sz); i++)
		{
			switch ((op>>16)&0x1f)
			{
			case 0: d[i] = s[i]; break; //vmov
			case 1: d[i] = fabs(s[i]); break; //vabs
			case 2: d[i] = -s[i]; break; //vneg
			case 4: if (s[i]<0) d[i]=0; else {if(s[i]>1) d[i] = 1; else d[i] = s[i];} break;
			case 5: if (s[i]<-1) d[i]=-1; else {if(s[i]>1) d[i] = 1; else d[i] = s[i];} break;
			case 16: d[i] = 1.0f/s[i]; break; //vrcp
			case 17: d[i] = 1.0f/sqrt(s[i]); break; //vrsq
			case 18: d[i] = sin((float)M_PI_2 * s[i]); break; //vsin
			case 19: d[i] = cos((float)M_PI_2 * s[i]); break; //vcos
			case 20: d[i] = powf(2.0f, s[i]); break;
			case 21: d[i] = log(s[i])/log(2.0f); break;
			//case 17: d[i] = 1.0f/sqrt(s[i]); break; //vrsq
			//case 17: d[i] = 1.0f/sqrt(s[i]); break; //vrsq
			case 22: d[i] = sqrt(s[i]); break; //vsqrt

			default:
				_dbg_assert_msg_(CPU,0,"Trying to interpret instruction that can't be interpreted");
				break;
			}
		}
		ApplyPrefixD(d,sz);
		WriteVector(d,sz,vd);
		PC+=4;
		EatPrefixes();
	}

	void Int_Vf2i(u32 op)
	{
		float s[4];
		int d[4];
		int vd = _VD;
		int vs = _VS;
		int imm = (op>>16)&0x1f;
		float mult = (float)(1<<imm);
		DataSize sz = GetVecSize(op);
		ReadVector(s,sz,vs);
		ApplySwizzleS(s,sz); //TODO: and the mask to kill everything but swizzle
		for (int i=0; i<GetNumElements(sz); i++)
		{
			switch ((op>>21)&0x1F)
			{
			case 16: d[i] = (int)floor(s[i]*mult+0.5f); break; //n
			case 17: d[i] = s[i]>=0?(int)floor(s[i]*mult):(int)ceil(s[i]*mult); break; //z
			case 18: d[i] = (int)ceil(s[i]*mult); break; //u
			case 19: d[i] = (int)floor(s[i]*mult); break; //d
			}
		}
		ApplyPrefixD((float*)d,sz); //TODO: and the mask to kill everything but mask
		WriteVector((float*)d,sz,vd);
		PC+=4;
		EatPrefixes();
	}

	void Int_Vi2f(u32 op)
	{
		int s[4];
		float d[4];
		int vd = _VD;
		int vs = _VS;
		int imm = (op>>16)&0x1f;
		float mult = 1.0f/(float)(1<<imm);
		DataSize sz = GetVecSize(op);
		ReadVector((float*)&s,sz,vs);
		ApplySwizzleS((float*)&s,sz); //TODO: and the mask to kill everything but swizzle
		for (int i=0; i<GetNumElements(sz); i++)
		{
			d[i] = (float)s[i] * mult;
		}
		ApplyPrefixD(d,sz); //TODO: and the mask to kill everything but mask
		WriteVector(d,sz,vd);
		PC+=4;
		EatPrefixes();
	}
	void Int_Vi2x(u32 op)
	{
		int s[4];
		u32 d[2];
		int vd = _VD;
		int vs = _VS;
		DataSize sz = GetVecSize(op);
		DataSize oz;
		int n = GetNumElements(sz);
		ReadVector((float*)s,sz,vs);
		ApplySwizzleS((float*)s,sz); //TODO: and the mask to kill everything but swizzle
		switch ((op>>16)&3)
		{
		case 0: //vi2uc
			{
				for (int i=0; i<4; i++)
				{
					int v = s[i];
					if (v<0) v=0;
					v>>=23;
					d[0] |= ((u32)v&0xFF)<<(i*8);
				}
				oz = V_Single;
			}
			break;
		case 1: //vi2c
			//break;
		case 2:  //vi2us
			//break;
		case 3:  //vi2s
			//break;
		default:
			_dbg_assert_msg_(CPU,0,"Trying to interpret instruction that can't be interpreted");
			break;
		}
		ApplyPrefixD((float*)d,oz);
		WriteVector((float*)d,oz,vd);
		PC+=4;
		EatPrefixes();
	}


	void Int_VDot(u32 op)
	{
		float s[4], t[4];
		float d;
		int vd = _VD;
		int vs = _VS;
		int vt = _VT;
		DataSize sz = GetVecSize(op);
		ReadVector(s,sz,vs);
		ApplySwizzleS(s,sz);
		ReadVector(t,sz,vt);
		ApplySwizzleT(t,sz);
		float sum=0;
		int n = GetNumElements(sz);
		for (int i=0; i<n; i++)
		{
			sum+=s[i]*t[i];
		}
		d = sum;
		ApplyPrefixD(&d,V_Single);
		V(vd) = d;
		PC+=4;
		EatPrefixes();
	}

	void Int_Vbfy(u32 op)
	{
		float s[4];
		float d[4];
		int vd = _VD;
		int vs = _VS;
		DataSize sz = GetVecSize(op);
		ReadVector(s,sz,vs);
		ApplySwizzleS(s,sz);
		int n = GetNumElements(sz);
		if (op & 0x10000)
		{
			//vbfy2
			d[0] = s[0] + s[2];
			d[1] = s[1] + s[3];
			d[2] = s[0] - s[2];
			d[3] = s[1] - s[3];
		}
		else
		{
			for (int i=0; i<n; i+=2)
			{
				d[i]   = s[i] + s[i+1];
				d[i+1] = s[i] - s[i+1];
			}
		}
		ApplyPrefixD(d,sz);
		WriteVector(d,sz,vd);
		PC+=4;
		EatPrefixes();
	}

	void Int_Vcrs(u32 op)
	{
		//half a cross product
		float s[4], t[4];
		float d[4];
		int vd = _VD;
		int vs = _VS;
		int vt = _VT;
		DataSize sz = GetVecSize(op);
		if (sz != V_Triple)
			_dbg_assert_msg_(CPU,0,"Trying to interpret instruction that can't be interpreted");

		ReadVector(s,sz,vs);
		ReadVector(t,sz,vt);
		// no swizzles allowed
		float sum=0;
		d[0] = s[1] * t[2];
		d[1] = s[2] * t[0];
		d[2] = s[0] * t[1];
		ApplyPrefixD(d,sz);
		WriteVector(d,sz,vd);
		PC+=4;
		EatPrefixes();
	}

	void Int_Vfad(u32 op)
	{
		float s[4];
		float d;
		int vd = _VD;
		int vs = _VS;
		DataSize sz = GetVecSize(op);
		ReadVector(s,sz,vs);
		ApplySwizzleS(s,sz);
		float sum=0;
		int n = GetNumElements(sz);
		for (int i=0; i<n; i++)
		{
			sum+=s[i];
		}
		d = sum;
		ApplyPrefixD(&d,V_Single);
		V(vd) = d;
		PC+=4;
		EatPrefixes();
	}


	void Int_VScl(u32 op)
	{
		float s[4];
		float d[4];
		int vd = _VD;
		int vs = _VS;
		int vt = _VT;
		DataSize sz = GetVecSize(op);
		ReadVector(s,sz,vs);
		ApplySwizzleS(s,sz);
		float scale = V(vt);
		int n = GetNumElements(sz);
		for (int i=0; i<n; i++)
		{
			d[i] = s[i]*scale;
		}
		ApplyPrefixD(d,sz);
		WriteVector(d,sz,vd);
		PC+=4;
		EatPrefixes();
	}


	void Int_Vrot(u32 op)
	{
		int vd = _VD;
		int vs = _VS;
		int imm = (op>>16) & 0x1f;
		DataSize sz = GetVecSize(op);
		int n = GetNumElements(sz);
		float angle = V(vs);
		bool negSin = (imm & 0x10) ? true : false;
		float sine = sinf(angle*(float)M_PI/2);
		if (negSin) sine=-sine;
		float cosine = cosf(angle*(float)M_PI/2);
		float d[4] = {0,0,0,0};
		if (((imm>>2)&3) == (imm&3))
		{
			for (int i=0; i<4; i++)
				d[i]=sine;
		}
		d[(imm>>2) & 3] = sine;
		d[imm&3] = cosine;
		WriteVector(d,sz,vd);
		PC+=4;
		EatPrefixes();
	}

	void Int_Vtfm(u32 op)
	{
		int vd = _VD;
		int vs = _VS;
		int vt = _VT;
		int ins = (op>>23) & 7;
		DataSize sz = GetVecSize(op);
		DataSize msz = GetMtxSize(op);
		int n = GetNumElements(sz);
		float s[16];
		ReadMatrix(s, msz, 4, vs);
		float t[4];
		ReadVector(s, sz, vt);
		float d[4];
		if (n == ins)
		{
			//homogenous
			for (int i=0; i<n; i++)
			{
				d[i] = 0.0f;
				for (int k=0; k<n; k++)
				{
					d[i] += (k==n-1) ? s[i*4+k] : (s[i+4*k] * t[k]);
				}
			}
		}
		else if (n == ins+1)
		{
			for (int i=0; i<n; i++)
			{
				d[i] = 0.0f;
				for (int k=0; k<n; k++)
				{
					d[i] += s[i+4*k] * t[k];
				}
			}
		}
		else
		{
			_dbg_assert_msg_(CPU,0,"Trying to interpret instruction that can't be interpreted (BADVTFM)");
		}
		WriteVector(d,sz,vd);
		PC+=4;
		EatPrefixes();
	}


	void Int_MatrixSet(u32 op)
	{
		int vd = op & 0x7F;
		int vs = (op >> 8) & 0x7f;

		switch ((op >> 16) & 15)
		{
		case 0: //vmmov
			//;break;
		case 3: //vmmidt
			//break;
		case 6: //vmzero
			//break;
		case 7: //vmone
			//break; 
		default:
			_dbg_assert_msg_(CPU,0,"Trying to interpret instruction that can't be interpreted");
			break;

		}
		PC += 4;
		EatPrefixes();
	}


	void Int_SV(u32 op)
	{
		s32 imm = (signed short)(op&0xFFFC);
		int vt = ((op>>16)&0x1f)|((op&3)<<5);
		int rs = (op>>21) & 0x1f;
		u32 addr = R(rs) + imm;

		switch (op>>26)
		{
		case 50: //lv.s
			{
				float f = ReadMemFloat(addr);
				//VI(vt) = ;
				WriteMatrix(&f,V_Single,4,vt);
			}
			break;
		case 58: //sv.s
			WriteMem32(addr, VI(vt));
			break;
		default:
			_dbg_assert_msg_(CPU,0,"Trying to interpret instruction that can't be interpreted");
			break;
		}
		PC+=4;
		//const char *name = MIPSGetName(op);
		//sprintf(out, "%s\t%s, %d(%s)",name,SN(vt,0),offset,RN(rs));
	}


	void Int_Mftv(u32 op)
	{
		int vr = op & 0xFF;
		int rt = _RT;
		switch ((op>>21)&0x1F)
		{
		case 3://mfv
			R(rt) = VI(vr);
			break;
		case 7: //mtv
			VI(vr) = R(rt);
			break;
		default:
			_dbg_assert_msg_(CPU,0,"Trying to interpret instruction that can't be interpreted");
			break;
		}
		PC+=4;
		//const char *name = MIPSGetName(op);
		//sprintf(out, "%s%s\t%s, %s",name,vr>127?"c":"", RN(rt), VN(vr, V_Single));
	}

#undef max

	void Int_Vcst(u32 op)
	{
		static const float constants[32] = 
		{
			0,
			std::numeric_limits<float>::max(),
			sqrt(2.0f),
			sqrt(0.5f),
			2.0f/sqrtf((float)M_PI),
			2.0f/(float)M_PI,
			1.0f/(float)M_PI,
			(float)M_PI/4,
			(float)M_PI/2,
			(float)M_PI,
			(float)M_E,
			(float)M_LOG2E,
			(float)M_LOG10E,
			(float)M_LN2,
			(float)M_LN10,
			2*(float)M_PI,
			(float)M_PI/6,
			0, //log(2), //"Log10(2)",
			0, //"Log2(10)",
			sqrt(3.0f)/2.0f, //"Sqrt(3)/2"
		};

		int conNum = (op>>16) & 0x1f;
		int vd = _VD;

		DataSize sz = GetVecSize(op);
		int numElems = GetNumElements(sz);
		float c = constants[conNum];
		float temp[4] = {c,c,c,c};
		WriteMatrix(temp,sz,4,vd);	
		PC+=4;
		EatPrefixes();
	}

	enum VCondition 
	{
		VC_FL,
		VC_EQ,
		VC_LT,
		VC_LE,
		VC_TR,
		VC_NE,
		VC_GE,
		VC_GT,
		VC_EZ,
		VC_EN,
		VC_EI,
		VC_ES,
		VC_NZ,
		VC_NN,
		VC_NI,
		VC_NS
	};

	void Int_Vcmp(u32 op)
	{
		int vt = _VT;
		int vs = _VS;
		int cond = op&15;
		DataSize sz = GetVecSize(op);
		int n = GetNumElements(sz);
		float s[4];
		float t[4];
		ReadVector(s,sz,vs);
		ApplySwizzleS(s,sz);
		ReadVector(t,sz,vt);
		ApplySwizzleT(t,sz);
		int cc = 0;
		int or = 0;
		int and = 1;
		for (int i=0; i<n; i++)
		{
			int c;
			switch (cond)
			{
			case VC_EZ: c = s[i] == 0.0f || s[i] == -0.0f; break;
			case VC_LT: c = s[i] < t[i]; break;
			case VC_LE: c = s[i] <= t[i]; break;
			case VC_TR: c = 1; break;
			case VC_FL: c = 0; break;
			case VC_NE: c = s[i] != t[i]; break;
			case VC_GT: c = s[i] > t[i]; break;
			case VC_GE: c = s[i] >= t[i]; break;
			case VC_NZ: c = s[i] != 0; break;
			default:
				_dbg_assert_msg_(CPU,0,"Unsupported vcmp condition code %i", cond);
				break;
			}
			cc|=(c<<i);
			or|=c;
			and&=c;
		}
		currentMIPS->vfpuCtrl[VFPU_CTRL_CC] = cc | (or<<4) | (and<<5);
		PC+=4;
		EatPrefixes();
	}


	void Int_Vcmov(u32 op)
	{
		int vs = _VS;
		int vd = _VD; 
		int tf = (op >> 19)&3;
		int imm3 = (op>>16)&7;
		DataSize sz = GetVecSize(op);
		int n = GetNumElements(sz);
		float s[4];
		float t[4];
		float d[4];
		ReadVector(s,sz,vs);
		ApplySwizzleS(s,sz);
		ReadVector(t,sz,vd); //Yes!
		ApplySwizzleT(t,sz);
		for (int i=0; i<n; i++)
			d[i] = t[i];

		int CC = currentMIPS->vfpuCtrl[VFPU_CTRL_CC];

		if (imm3<6)
		{
			if (((CC>>imm3)&1) == !tf)
			{
				for (int i=0; i<n; i++)
					d[i] = s[i];
			}
		}
		else if (imm3 == 6)
		{
			for (int i=0; i<n; i++)
			{
				if (((CC>>i)&1)  == !tf)
					d[i] = s[i];
			}
		}
		else 
		{
			_dbg_assert_msg_(CPU,0,"Bad Imm3 in cmov");
		}
		ApplyPrefixD(d, sz);
		WriteVector(d, sz, vd);
		PC+=4;
		EatPrefixes();
	}

	void Int_VecDo3(u32 op)
	{
		int vd = _VD;
		int vs = _VS;
		int vt = _VT;
		DataSize sz = GetVecSize(op);
		float s[4];
		float t[4];
		float d[4];
		ReadVector(s,sz,vs);
		ApplySwizzleS(s,sz);
		ReadVector(t,sz,vt);
		ApplySwizzleT(t,sz);
		for (int i=0; i<GetNumElements(sz); i++)
		{
			switch(op>>26)
			{
			case 24: //VFPU0
				switch ((op>>23)&7)
				{
				case 0: d[i] = s[i] + t[i]; break; //vadd
				case 1: d[i] = s[i] - t[i]; break; //vsub
				case 7: d[i] = s[i] / t[i]; break; //vdiv
				default: goto bad;
				}
				break;
			case 25: //VFPU1
				switch ((op>>23)&7)
				{
				case 0: d[i] = s[i] * t[i]; break; //vmul
				default: goto bad;
				}
				break;
				/*
			case 27: //VFPU3
				switch ((op>>23)&7)
				{
				case 0: d[i] = s[i] * t[i]; break; //vmul
				}
				break;*/
			default:
bad:
				_dbg_assert_msg_(CPU,0,"Trying to interpret instruction that can't be interpreted");
				break;
			}
		}
		ApplyPrefixD(d,sz);
		WriteVector(d,sz,vd);
		PC+=4;
		EatPrefixes();
	}


	void Int_CrossQuat(u32 op)
	{
		int vd = _VD;
		int vs = _VS;
		int vt = _VT;
		DataSize sz = GetVecSize(op);
		float s[4];
		float t[4];
		float d[4];
		ReadVector(s,sz,vs);
		ReadVector(t,sz,vt);
		switch (sz)
		{
		case V_Triple:
			d[0] = s[1]*t[2] - s[2]*t[1];
			d[1] = s[2]*t[0] - s[0]*t[2];
			d[2] = s[0]*t[1] - s[1]*t[0];
			//cross
			break;
		case V_Quad:
			//quat
			break;
		default:
			_dbg_assert_msg_(CPU,0,"Trying to interpret instruction that can't be interpreted");
			break;
		}
		WriteVector(d,sz,vd);
		PC+=4;
		EatPrefixes();
	}

}
