#include "stdafx.h"
#include "../PSPFirmwareHLE.h"
#include "../../MIPS/MIPS.h"

#include "sceKernel.h"
#include "sceUtility.h"

#include "../../EmuThread.h"



namespace PSPHLE
{

	enum SceUtilitySavedataType
	{
		SCE_UTILITY_SAVEDATA_TYPE_AUTOLOAD		= 0,
		SCE_UTILITY_SAVEDATA_TYPE_AUTOSAVE		= 1,
		SCE_UTILITY_SAVEDATA_TYPE_LOAD			= 2,
		SCE_UTILITY_SAVEDATA_TYPE_SAVE			= 3,
		SCE_UTILITY_SAVEDATA_TYPE_LISTLOAD		= 4,
		SCE_UTILITY_SAVEDATA_TYPE_LISTSAVE		= 5,
		SCE_UTILITY_SAVEDATA_TYPE_LISTDELETE	= 6,
		SCE_UTILITY_SAVEDATA_TYPE_DELETE		= 7,
		SCE_UTILITY_SAVEDATA_TYPE_SIZES			= 8	
	} ;

#define SCE_UTILITY_SAVEDATA_ERROR_TYPE					(0x80110300)

#define SCE_UTILITY_SAVEDATA_ERROR_LOAD_NO_MS			(0x80110301)
#define SCE_UTILITY_SAVEDATA_ERROR_LOAD_EJECT_MS		(0x80110302)
#define SCE_UTILITY_SAVEDATA_ERROR_LOAD_ACCESS_ERROR	(0x80110305)
#define SCE_UTILITY_SAVEDATA_ERROR_LOAD_DATA_BROKEN		(0x80110306)
#define SCE_UTILITY_SAVEDATA_ERROR_LOAD_NO_DATA			(0x80110307)
#define SCE_UTILITY_SAVEDATA_ERROR_LOAD_PARAM			(0x80110308)
#define SCE_UTILITY_SAVEDATA_ERROR_LOAD_INTERNAL		(0x8011030b)

#define SCE_UTILITY_SAVEDATA_ERROR_SAVE_NO_MS			(0x80110381)
#define SCE_UTILITY_SAVEDATA_ERROR_SAVE_EJECT_MS		(0x80110382)
#define SCE_UTILITY_SAVEDATA_ERROR_SAVE_MS_NOSPACE		(0x80110383)
#define SCE_UTILITY_SAVEDATA_ERROR_SAVE_MS_PROTECTED	(0x80110384)
#define SCE_UTILITY_SAVEDATA_ERROR_SAVE_ACCESS_ERROR	(0x80110385)
#define SCE_UTILITY_SAVEDATA_ERROR_SAVE_PARAM			(0x80110388)
#define SCE_UTILITY_SAVEDATA_ERROR_SAVE_NO_UMD			(0x80110389)
#define SCE_UTILITY_SAVEDATA_ERROR_SAVE_WRONG_UMD		(0x8011038a)
#define SCE_UTILITY_SAVEDATA_ERROR_SAVE_INTERNAL		(0x8011038b)

#define SCE_UTILITY_SAVEDATA_ERROR_DELETE_NO_MS			(0x80110341)
#define SCE_UTILITY_SAVEDATA_ERROR_DELETE_EJECT_MS		(0x80110342)
#define SCE_UTILITY_SAVEDATA_ERROR_DELETE_MS_PROTECTED	(0x80110344)
#define SCE_UTILITY_SAVEDATA_ERROR_DELETE_ACCESS_ERROR	(0x80110345)
#define SCE_UTILITY_SAVEDATA_ERROR_DELETE_NO_DATA		(0x80110347)
#define SCE_UTILITY_SAVEDATA_ERROR_DELETE_PARAM			(0x80110348)
#define SCE_UTILITY_SAVEDATA_ERROR_DELETE_INTERNAL		(0x8011034b)

#define SCE_UTILITY_SAVEDATA_ERROR_SIZES_NO_MS			(0x801103C1)
#define SCE_UTILITY_SAVEDATA_ERROR_SIZES_EJECT_MS		(0x801103C2)
#define SCE_UTILITY_SAVEDATA_ERROR_SIZES_ACCESS_ERROR	(0x801103C5)
#define SCE_UTILITY_SAVEDATA_ERROR_SIZES_NO_DATA		(0x801103C7)
#define SCE_UTILITY_SAVEDATA_ERROR_SIZES_PARAM			(0x801103C8)
#define SCE_UTILITY_SAVEDATA_ERROR_SIZES_NO_UMD			(0x801103C9)
#define SCE_UTILITY_SAVEDATA_ERROR_SIZES_WRONG_UMD		(0x801103Ca)
#define SCE_UTILITY_SAVEDATA_ERROR_SIZES_INTERNAL		(0x801103Cb)

#define SCE_UTILITY_STATUS_NONE 		0
#define SCE_UTILITY_STATUS_INITIALIZE	1
#define SCE_UTILITY_STATUS_RUNNING 		2
#define SCE_UTILITY_STATUS_FINISHED 	3
#define SCE_UTILITY_STATUS_SHUTDOWN 	4


	/** title, savedataTitle, detail: parts of the unencrypted SFO
	data, it contains what the VSH and standard load screen shows */
	typedef struct PspUtilitySavedataSFOParam
	{
		char title[0x80];
		char savedataTitle[0x80];
		char detail[0x400];
		unsigned char parentalLevel;
		unsigned char unknown[3];
	} PspUtilitySavedataSFOParam;

	typedef struct PspUtilitySavedataFileData {
		void *buf;
		SceSize bufSize;
		SceSize size;	/* ??? - why are there two sizes? */
		int unknown;
	} PspUtilitySavedataFileData;

	/** Structure to hold the parameters for the ::sceUtilitySavedataInitStart function.
	*/
	typedef struct SceUtilitySavedataParam
	{
		/** Size of the structure */
		SceSize size;

		int language;

		int buttonSwap;

		int unknown[4];
		int result;
		int unknown2[4];

		/** mode: 0 to load, 1 to save */
		int mode;
		int bind;

		/** unknown13 use 0x10 */
		int overwriteMode;

		/** gameName: name used from the game for saves, equal for all saves */
		char gameName[16];
		/** saveName: name of the particular save, normally a number */
		char saveName[24];
		/** fileName: name of the data file of the game for example DATA.BIN */
		char fileName[16];

		/** pointer to a buffer that will contain data file unencrypted data */
		void *dataBuf;
		/** size of allocated space to dataBuf */
		SceSize dataBufSize;
		SceSize dataSize;

		PspUtilitySavedataSFOParam sfoParam;

		PspUtilitySavedataFileData icon0FileData;
		PspUtilitySavedataFileData icon1FileData;
		PspUtilitySavedataFileData pic1FileData;
		PspUtilitySavedataFileData snd0FileData;

		unsigned char unknown17[4];
	} SceUtilitySavedataParam;


	struct UtilityState
	{
		int status;
	};

	static UtilityState state;

	void sceUtilitySavedataGetStatus()
	{
		LOG(HLE,"%i=sceUtilitySavedataGetStatus()", state.status);
		RETURN(state.status);
		if (state.status == SCE_UTILITY_STATUS_SHUTDOWN)
			state.status = SCE_UTILITY_STATUS_NONE;
	}

	void sceUtilitySavedataShutdownStart()
	{
		LOG(HLE,"sceUtilitySavedataShutdownStart()");
		state.status = SCE_UTILITY_STATUS_SHUTDOWN;
		RETURN(0);
	}

	void sceUtilitySavedataUpdate()
	{
		LOG(HLE,"sceUtilitySavedataUpdate");
		//draw savedata UI here
		if (state.status == SCE_UTILITY_STATUS_RUNNING)
			state.status = SCE_UTILITY_STATUS_FINISHED;

		if (state.status == SCE_UTILITY_STATUS_SHUTDOWN)
			state.status = SCE_UTILITY_STATUS_NONE;
		RETURN(0);
	}
	

	void sceUtilitySavedataInitStart()
	{
		SceUtilitySavedataParam *param = (SceUtilitySavedataParam*)GetMemPointer(PARAM(0));

		LOG(HLE,"sceUtilitySavedataInitStart(%08x)", PARAM(0));
		
		if (param->mode == 0) //load
		{
			RETURN(SCE_UTILITY_SAVEDATA_ERROR_LOAD_NO_DATA);
			LOG(HLE, "Loading. Title: %s Save: %s File: %s", param->gameName, param->saveName, param->fileName);
		}
		else 
		{
			//save
			LOG(HLE, "Saving. Title: %s Save: %s File: %s", param->gameName, param->saveName, param->fileName);
			RETURN(SCE_UTILITY_SAVEDATA_ERROR_SAVE_NO_MS);
		}
		
		state.status = SCE_UTILITY_STATUS_RUNNING;
	}

	void sceUtilityGetSystemParamInt()
	{
		LOG(HLE,"sceUtilityGetSystemParamInt(%i, %08x)", PARAM(0), PARAM(1));
		u32 *outPtr = (u32*)GetMemPointer(PARAM(1));
		int defaultValues[16] = 
		{
			0,
			0,
			0,
			0,//date notation
			0,//time notation
			0,//timezone offset in minutes
			0,//daylight savings
			1,//language (0=jap 1=eng)
			0,
		};
		*outPtr = defaultValues[PARAM(0)];
		RETURN(0);
	}
}
