#include "stdafx.h"

#include "PSPFirmwareHLE.h"
#include "../MemMap.h"

//#include "../MIPS/MIPS.h"

#include "HLE/HLETables.h"
#include "PSPSystem.h"
#include "HLE/sceIo.h"
#include "HLE/sceAudio.h"
#include "HLE/sceKernelMEmory.h"


namespace PSPHLE
{

	//unused
	struct PSPHeader
	{
		char psp[4];
		u32 version; //00080000
		short whatever; //0101
		char sometext[28];
		short whatever2; //0102
		u32 filesizedecrypted;
		u32 filesize;
		u32 unknownoffsets[3];
		short whatever3[2]; //0x40 0x40
		u32 whatever4[2]; //00
		u32 whatever5; //1067008
		u32 whatever6[2]; //00
		u32 whatever7; //1067008
		u32 whatever8; //2553296
		u32 whatever9[8]; //0
		u32 whatever10; //12
		u32 encryptedShit[12];
		u32 filesizedecrypted2;
		u32 whatever11;//0x80
		u32 whatever12[6];
	};

	int GetModuleIndex(const char *moduleName)
	{
		//todo: use map or something faster than linear search
		for (int i=0; i<numModules; i++)
			if (strcmp(moduleName, modules[i].name) == 0)
				return i;
		return -1;
	}

	int GetFuncIndex(int moduleIndex, u32 nib)
	{
		const HLEModule &module = modules[moduleIndex];
		for (int i=0; i<module.numFunctions; i++)
		{
			if (module.funcTable[i].ID == FID(nib))
				return i;
		}
		return -1;
	}

	const HLEFunction *GetFunc(const char *moduleName, u32 nib)
	{
		int moduleIndex = PSPHLE::GetModuleIndex(moduleName);
		if (moduleIndex != -1)
		{
			int idx = PSPHLE::GetFuncIndex(moduleIndex, nib);
			if (idx != -1)
				return &(modules[moduleIndex].funcTable[idx]);
		}
		return 0;
	}

	const char *GetFuncName(const char *moduleName, u32 nib)
	{
		const HLEFunction *func = GetFunc(moduleName,nib);
		if (func)
			return func->name;
		else
		{
			static char temp[256];
			sprintf(temp,"[UNK:%08x]",nib);
			return temp;
		}
	}

	const char *GetModuleName(int moduleindex)
	{
		return modules[moduleindex].name;
	}

	void WriteSyscall(const char *moduleName, u32 nib, u32 address)
	{
		if (nib == 0)
		{
			WriteMem32(address+4, 0); //patched out?
			return;
		}
		int modindex = PSPHLE::GetModuleIndex(moduleName);
		if (modindex != -1)
		{
			int funcindex = PSPHLE::GetFuncIndex(modindex, nib);
			if (funcindex != -1)
			{
				WriteMem32(address, 0x0000000c | (modindex<<18) | (funcindex<<6));
			}
			else
			{
				WriteMem32(address, 0x0003FFCC | (modindex<<18)); //invalid syscall
			}
			WriteMem32(address+4, 0); //nop
		}
		else
		{
			LOG(HLE, "Unknown module %s!", moduleName);
			WriteMem32(address+4, 0x0003FFCC); //invalid syscall
		}
	}


	const char *GetFuncName(int moduleIndex, int func)
	{
		if (moduleIndex>=0 && moduleIndex<=numModules)
		{
			const HLEModule &module = modules[moduleIndex];
			if (func>=0 && func <= module.numFunctions)
			{
				return module.funcTable[func].name;
			}
		}
		return "[unknown]";
	}

	void CallSyscall(u32 op)
	{
		u32 callno = (op>>6) & 0xFFFFF; //20 bits
		int funcnum = callno & 0xFFF;
		int modulenum = (callno & 0xFF000) >> 12;
		if (funcnum == 0xfff)
		{
			_dbg_assert_msg_(HLE,0,"Unknown syscall");
			LOG(HLE,"Unknown syscall: Module: %s", modules[modulenum].name); 
			return;
		}
		HLEFunc func = modules[modulenum].funcTable[funcnum].func;
		if (func)
		{
			func();
		}
		else
		{
			LOG(HLE,"Unimplemented HLE function %s",modules[modulenum].funcTable[funcnum].name);
		}
/*
		__asm
		{
			mov ecx,numparams
			mov edi,rega0
			loop:
			mov eax,[edi]
			add edi,4
			push eax
			dec ecx
			jnz loop
			call func
		}*/
	}
}
