
#pragma once

#include "../../Globals.h"

#include <vector>
#include <list>


// Generic allocator thingy
// Allocates blocks from a range

class BlockAllocator
{
	struct Block
	{
		Block(u32 _start, u32 _size, bool _taken) : start(_start), size(_size), taken(_taken) 
		{
			strncpy_s(tag, "Empty", _TRUNCATE);
		}
		void SetTag(const char *_tag) {
			if (_tag)
				strncpy_s(tag, _tag, _TRUNCATE);
			else
				strncpy_s(tag, "---", _TRUNCATE);
		}
		bool taken;
		u32 start;
		u32 size;
		char tag[16];
	};

	std::list<Block> blocks;

	Block *GetBlockFromAddress(u32 addr);
	std::list<Block>::iterator GetBlockIterFromAddress(u32 addr);

public:
	BlockAllocator();
	~BlockAllocator();

	void Init(u32 _rangeStart, u32 _rangeSize);
	void Shutdown();

	void ListBlocks();

	u32 Alloc(u32 size, bool fromEnd = false, const char *tag = 0);
	u32 AllocAt(u32 position, u32 size, const char *tag = 0);

	void Free(u32 position);
	bool IsBlockFree(u32 position) {
		Block *b;
		if (b = GetBlockFromAddress(position))
			return !b->taken;
		else
			return false;
	}

	void MergeFreeBlocks();

	u32 GetBlockStartFromAddress(u32 addr);
	u32 GetBlockSizeFromAddress(u32 addr);
	u32 GetLargestFreeBlockSize();
	u32 GetTotalFreeBytes();
};